/* file:   grey.c
 * author: Johannes Hubertz
 * date:   20090804 (initial)
 * update: 20110627
 * license: GPL version 2 or later
 */
#if JH_GREY

#include <syslog.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "smtpd.h"

#define  LINE_LEN 200
#define  LINE_USE LINE_LEN - 1

time_t aktuell;
/* fixed filenames for now */
char *wantspam  = "/etc/immediate-list";
char *whitefrom = "/etc/senders-whitened";
char *blackfrom = "/etc/senders-blackend";
char *whitefile = "/etc/white-hosts";  
char *blackfile = "/etc/black-hosts";
extern int jh_grey_min_time;

unsigned long line_ctr;
static char found[LINE_LEN + 1]; /* global safe for found value   */
static char zeile[LINE_LEN + 1]; /* filecontent is read here      */
static char grey_fname[30]; /* f.e:/etc/lists/192/grey-hosts\0    */
/*                                 12345678901234567890123456     */

/*
 *
 * Black and Whitelisting is now done by ip_addr only 
 *                        (this might be changed in future)
 * first done 20080109
 *
 * is_in_file simply checks if given string ip_addr is found at 
 * beginning of a text-line in file filename
 *       so you can have a comment behind every ip, if you like
 *
 * returns: 0 if not in file, else 1
 */
int is_in_file(char * filename, char * ip_addr)
{
	FILE *fp;
	int len;
	line_ctr=0;
	if (NULL == (fp = fopen(filename,"r"))){
	    /* always complain about lazy admin who did not create file */
            syslog(LOG_DEBUG, "does not exist: %s", filename);
	    return (0); /* IP not found */
	}
	/* the file is open now */
	while (!feof(fp)){
		bzero(&zeile[0], LINE_USE );
		bzero(&found[0], LINE_USE );
		line_ctr++;
		if (fgets (zeile, LINE_USE, fp)) {
		    /* use the length from ip-number in file,            */
                    /* f.e. 10.0 might be enough for 10.0/16             */ 
                    /*      but remind it is stringcompare,              */
                    /*      no number compare !!!                        */ 
		    /* CIDR is much more complicated, may be done later  */
		    /* first char may also be a nonnumeric and not dot,  */
                    /*      so we are able to have comment lines as well */

		    len = 0; /* every line starts at the first position  */
		    /* file might have been corrupted? check linelimits! */
		    while (( (zeile[len] >= '0'  && zeile[len ] <= '9')|| \
                              zeile[len] == '.') && len < LINE_LEN)   \
			   len++;
		    /* if len == 0 we have a comment found, proceed      */ 
		    if ((len > 0) && (!strncmp(zeile,ip_addr,len))) {
			/* found him */
			fclose(fp);
			strncpy(found,zeile,LINE_USE);
			found[LINE_USE] = (char) 0;
			return (1); /* IP found */
		    }
		}
	}
	fclose(fp);
	return (0); /* IP not found */
}

/* 
 * make_grey_filename takes up to the first three chars of 
 * ip_addr and appends them to a consant string first copied 
 * to global var grey_fname then appends constant name 
 * grey-hosts, global var grey_fname holds result.
 */

int make_grey_filename(char * ip_addr)
{
    int z = 0;
    char this_ip[20]; /* first bytes of ip as string to build to*/
    /* now first octet only (decimal number) is appended */
    /* any later version may use more than the first one */
    while (ip_addr[z] >= '0' && ip_addr[z] <= '9') {
	this_ip[z]   = ip_addr[z]; 
	this_ip[++z] = (char) 0;
    }
    strcpy(grey_fname, "/etc/lists/");
    strcat(grey_fname, this_ip);
    strcat(grey_fname, "/grey-hosts");
    return 0;
}

/*
 * append_to_grey appends record to greylistfile:
 * Params: filename, email from, email to.
 * also uses global parameter aktuell (timestamp)
 *
 */

int append_to_grey(char *this_file, char *ip_addr, char *from, char *to){
    FILE *fp;
    if (NULL == (fp = fopen(this_file,"a"))){
        syslog(LOG_CRIT, "file not writable: %s, abort", grey_fname);
        abort();
    } 
    /* the file is open now for appending */
    fprintf(fp, "%-20s%ld  %s, %s\n", ip_addr, aktuell, from, to);
    fclose(fp);
    syslog(LOG_DEBUG, "%s greylist:entered:-", ip_addr);
    return (2); /* ip_address entered into file, soft NOT OK */
}

/* is_in_grey_file always and only looks from to to first match
 * in grey-hosts.
 * In this version, always appends quadrupel as well to achive
 * a somehow autowhitelisting effect for frequent mailers
 * This feature is done by running expire.pl periodically on 
 * the grey-list-files as this removes old entries from the top
 * If you do not like this feature, just comment the line 
 */

int is_in_grey_file(char *ip_addr, char *from, char *to){
    FILE *fp;
    char fline[LINE_LEN]; /* line from file */
    char nline[LINE_LEN]; /* line from network-parameters */
    int z, n_len;
    unsigned long timestamp, diff, jh_grey_min;
    jh_grey_min = (long) jh_grey_min_time;
    make_grey_filename(ip_addr);
    if (NULL == (fp = fopen(grey_fname,"r"))){
        /* file unreadable, does not exist? */
        syslog(LOG_DEBUG, "creating: %s", grey_fname);
        return append_to_grey(grey_fname, ip_addr, from, to);
    }
    bzero(&fline[0], LINE_USE );
    sprintf(nline, "%-20s%ld  %s, %s\n", ip_addr, aktuell, from, to);
    nline[LINE_USE] = (char) 0;
    n_len = strlen(nline);
    /* the file is open now for reading, search his ip and when  */
    while (!feof(fp)){
	bzero(&fline[0], LINE_USE );
        if (fgets (fline, LINE_USE, fp)) {
	    /* copy timestamp from fline into nline for compare success */
	    /*      timestamp is to be compared later                   */
	    for (z=20; z<30; z++) nline[z] = fline[z];
	    if (!strncmp(fline, nline, n_len)) {
		/* we found him */
		fclose(fp);
		timestamp=atol(&fline[20]);
		diff = aktuell - timestamp;
                syslog(LOG_DEBUG, "%s greylist:%ld:since:%ld:diff:%ld:min:%ld", \
		   ip_addr, aktuell, timestamp, diff, jh_grey_min);
		/*if (diff >= jh_grey_min) { v 2.0-9 changed to following line on 20100902 */
		if ((diff >= jh_grey_min) || (jh_grey_min == 0)) {
		    /* always append new record on success */
                    append_to_grey(grey_fname, ip_addr, from, to); 
	            syslog(LOG_DEBUG, "%s greylist:%ld needed, has:%ld:+", \
			ip_addr, jh_grey_min, diff);
		    return (1); /* soft OK */
		} else {
	            syslog(LOG_DEBUG, "%s greylist:%ld needed, has:%ld:-", \
			ip_addr, jh_grey_min, diff);
		    return (2); /* soft NOT OK */
		}
	    } /* strncmp: not equal, try next line from file */
	}
    }
    fclose(fp);
    /* info was not on file, so result is soft NOT OK and store them */
    return append_to_grey(grey_fname, ip_addr, from, to); 
}

/* find short string (pattern from file) in a long string search (online)  */
int find_pattern_in_string_at_its_end(char *pattern, char *search) {
    int pat_len, sea_len, i, pos;
    char swap[LINE_LEN + 1];
    /*  1234567890123456789012345
     *  string_1_is_short         ==> len = 17
     *  string_2_is_much_longer   ==> len = 23
     *  pos = 23 - 17 = 6 ==> compare from the beginning of string 1 
     *        string_1_is_short   ==> len = 17
     *  string_2_is_much_longer   ==> len = 23
     *  diffs on the first character ==> quick ;-)
     */
    pat_len = strlen(pattern); 
    sea_len = strlen(search);
    pos = sea_len - pat_len;
    if (pos < 0) {
	strcpy(swap,search);
	strcpy(search,pattern);
	strcpy(pattern,swap);
    }
    return strcmp(&search[pos],pattern);
} /* find_pattern_in_string_at_its_end */

/* wants_from: function searches file for sender-address
 *             compares with actual from. 
 *             if found, acts like white-listet
 * fixed:      filenames, see global variables 
 * returns:    0 if not found
 *             0 if file not found, which is not any problem for now
 *             1 if address is found in file, mentioned in syslog,
*/
int wants_from(char *fname, char *from) {
    FILE *fp;
    int at_pos, ctr, len;
    line_ctr = 0;
    bzero(&zeile[0], LINE_USE);
    bzero(&found[0], LINE_USE);
    ctr = 0;
    len = 0;
    at_pos = 0;
    /* prelimaries: tolower first, might have braces and capital letters online */
    strcpy(zeile, from);
    while (zeile[len] > ' ' && len < LINE_LEN) {
        if (zeile[len] != '<' && zeile[len] != '>') {
            found[ctr++] = tolower(zeile[len]);
        }
        len++;
    }
    /* emails from-address is tolower and in variable: found */
    if (NULL == (fp = fopen(fname,"r"))) {
        return (0); /* file not found, no problem! */
    }
    /* the file is open now, read until found or EoF */
    /* syslog(LOG_DEBUG, "file %-15s is searched for:%s:", \
        fname, found); */
    while (!feof(fp)){
        bzero(&zeile[0], LINE_USE);
        if (fgets (zeile, LINE_USE, fp)) {
            line_ctr++;
            zeile[strlen(zeile) - 1] = (char) 0;
            /* no leading blanks allowed! */
            len=0;
            while (zeile[len] > ' ' && len < LINE_LEN ) {
                zeile[len] = tolower(zeile[len]);
                /* comment is allowed at every position in every line */
                if (zeile[len] == '#') len = LINE_USE;
                len++;
            }
            /* syslog(LOG_DEBUG, "look %-15s: line %03ld:%s: cmp :%s:", \
                fname, line_ctr, zeile, found); */
            /* one line read from file and tolower now, len indicates usability */
	    if (( len > 0 ) && (len < LINE_LEN)) {
                if (!find_pattern_in_string_at_its_end(zeile, found)) {
		    fclose(fp);
		    syslog(LOG_DEBUG, "look %-15s: line %03ld:%s: equ :%s:", \
			fname, line_ctr, zeile, found);
		    return (1); /* found */ 
                }
            }
        } /* fgets */
    } /* while */
    fclose(fp);
    return (0);
} /* wants_from */

/* wants_spam: funtion reads file for receipient-address within.
 *             added 20110623, JH
 * parameters: one: char *to, which points to receipient address online
 * fixed:      filename, see global variable wantspam
 * returns:    0 if not found
 *             0 if file not found, which is not any problem for now
 *             1 if address is found in file, mentioned in syslog,
*/
int wants_spam(char *to) {
    FILE *fp;
    int ctr, len;
    line_ctr = 0;
    bzero(&zeile[0], LINE_USE);
    bzero(&found[0], LINE_USE);
    ctr = 0;
    len = 0;
    /* prelimaries: tolower first, might have braces and capital letters online */
    strcpy(zeile, to);
    while (zeile[len] > ' ' && len < LINE_LEN) {
	if (zeile[len] != '<' && zeile[len] != '>') {
	    found[ctr++] = tolower(zeile[len]);
	}
	len++;
    }
    /* to is now copied to found, but lower and no braces */
    /* we have a fixed filname for now in wantspam */
    if (NULL == (fp = fopen(wantspam,"r"))) {
	return (0); /* file not found, no problem! */
    }
    /* the file is open now, read until found or EoF */
    /* syslog(LOG_DEBUG, "file %-15s is searched for:%s:", \
	wantspam, found); */
    while (!feof(fp)){
	bzero(&zeile[0], LINE_USE);
	if (fgets (zeile, LINE_USE, fp)) {
	    line_ctr++;
	    zeile[strlen(zeile) - 1] = (char) 0;
	    /* syslog(LOG_DEBUG, "look %-15s: line %03ld:%s: cmp :%s:", \
		wantspam, line_ctr, zeile, found); */
	    /* no leading blanks allowed! */
	    len=0;
	    while (zeile[len] > ' ' && len < LINE_LEN ) {
		zeile[len] = tolower(zeile[len]);
		/* comment is allowed at every position in every line */
		if (zeile[len] == '#') len = LINE_USE;
		len++;
	    }
	    /* one line read from file and tolower now, len indicates usability */
	    if (( len > 0 ) && (len < LINE_LEN)) {
		if ((len > 0) && (!strncmp(zeile,found,len))) {
		    /* found him, close file, syslog */
		    fclose(fp);
		    syslog(LOG_DEBUG, "look %-15s: line %03ld:%s: equ :%s:", \
			wantspam, line_ctr, zeile, found);
		    return (1); /* found */ 
		}
	    }
	} /* fgets */
    } /* while */
    fclose(fp);
    return (0); /* not found */
} /* wants_spam */

/* check_grey: main function
 * parameters: 3, char *ipaddr, char *from, char *to
 *             ipaddr is string with IPaddress of actual sender
 *             from is string with actual sender email address
 *             to is string with actual receipient email address
 * returns:    0 (hard OK)     (whitelist, wants_spam)
 *             1 (soft OK)     (greylist after wait-time)
 *             2 (soft NOT OK) (greylist befor wait-time)
 *             3 (hard NOT OK) (blacklist)
 * remarks:    whitelist  (ipaddress) suppresses all other criteria
 *             blacklist  (ipaddress) suppresses greylist and wants_*
 *             wants_from (emailaddr) suppresses greylist
 *             wants_spam (emailaddr) suppresses greylist
*/
int check_grey(char *ip_addr, char *from, char *to) {
  /* first get acutal time here */
  time(&aktuell);
  if (is_in_file(whitefile,ip_addr)){
        syslog(LOG_DEBUG, "white-listed in line %03ld: %-15s:%s", \
		line_ctr, ip_addr, &found[0]);
        return (0); /* hard OK */
  } else if (is_in_file(blackfile,ip_addr)){
        syslog(LOG_DEBUG, "black-listed in line %03ld: %s", \
		line_ctr, ip_addr);
        return (3); /* hard NOT OK */
  }
  if (wants_from(whitefrom,from)) return (0);	/* hard     OK because of senders address is on file */
  if (wants_from(blackfrom,from)) return (3);	/* hard NOT OK because of senders address is on file */
  if (wants_spam(to))             return (0);	/* hard     OK because of receipient address is on file */
  return is_in_grey_file(ip_addr, from, to); /* soft states */
} /* check_grey */

#endif /* JH_GREY */
